@@ -25,6 +25,7 @@ THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 """
 
 import enum
+import functools
 import itertools
 from itertools import zip_longest
 from .itertools import PushbackIterator
@@ -45,7 +46,7 @@ def _export(o):
 
 
 @_export
-def re_partition(s, pattern, count=1, *, flags=0):
+def re_partition(s, pattern, count=1, *, flags=0, reverse=False):
     """
     Like str.partition, but pattern is matched as a regular expression.
 
@@ -77,7 +78,12 @@ def re_partition(s, pattern, count=1, *, flags=0):
 
     If pattern is a string or bytes, flags is passed in
     as the flags argument to re.compile.
+
+    If reverse is true, partitions starting at the right,
+    like re_rpartition.
     """
+    if reverse:
+        return re_rpartition(s, pattern, count, flags=flags)
     if not isinstance(pattern, re.Pattern):
         pattern = re.compile(pattern, flags=flags)
     if not (isinstance(s, (str, bytes)) and
@@ -184,7 +190,7 @@ def re_rpartition(s, pattern, count=1, *, flags=0):
 
 # a list of all unicode whitespace characters
 _export_name('whitespace')
-whitespace = [
+whitespace = (
     '\t'    , #     9 0x0009 - tab
     '\n'    , #    10 0x000a - newline
     '\x0b'  , #    11 0x000b - vertical tab
@@ -216,14 +222,14 @@ whitespace = [
     '\u3000', # 12288 0x3000 - ideographic space
 
     '\r\n'  , # the classic DOS newline sequence
-    ]
+    )
 
 # this omits the DOS convention '\r\n'
 _export_name('whitespace_without_dos')
-whitespace_without_dos = [s for s in whitespace if s != b'\r\n']
+whitespace_without_dos = tuple(s for s in whitespace if s != '\r\n')
 
 _export_name('newlines')
-newlines = [
+newlines = (
     '\n'    , #   10 0x000a - newline
     '\x0b'  , #   11 0x000b - vertical tab
     '\x0c'  , #   12 0x000c - form feed
@@ -239,11 +245,11 @@ newlines = [
 
     # '\n\r' sorry, Acorn and RISC OS, you have to add this yourselves.
     # I'm worried this would cause bugs with a malformed DOS newline.
-    ]
+    )
 
 # this omits the DOS convention '\r\n'
 _export_name('newlines_without_dos')
-newlines_without_dos = [s for s in newlines if s != b'\r\n']
+newlines_without_dos = tuple(s for s in newlines if s != '\r\n')
 
 def _cheap_encode_iterable_of_strings(iterable, encoding='ascii'):
     a = []
@@ -268,7 +274,7 @@ ascii_whitespace_without_dos = [s for s in ascii_whitespace if s != b'\r\n']
 _export_name('ascii_newlines')
 ascii_newlines   = _cheap_encode_iterable_of_strings(newlines,   "ascii")
 _export_name('ascii_newlines_without_dos')
-ascii_newlines_without_dos = [s for s in ascii_newlines if s != b'\r\n']
+ascii_newlines_without_dos = tuple(s for s in ascii_newlines if s != b'\r\n')
 
 # use these with multisplit if your bytes strings are encoded as utf-8.
 _export_name('utf8_whitespace')
@@ -278,7 +284,7 @@ utf8_whitespace_without_dos = [s for s in utf8_whitespace if s != b'\r\n']
 _export_name('utf8_newlines')
 utf8_newlines   = _cheap_encode_iterable_of_strings(newlines,   "utf-8")
 _export_name('utf8_newlines_without_dos')
-utf8_newlines_without_dos =  [s for s in utf8_newlines if s != b'\r\n']
+utf8_newlines_without_dos =  tuple(s for s in utf8_newlines if s != b'\r\n')
 
 
 def _re_quote(s):
@@ -296,7 +302,9 @@ def _re_quote(s):
             s = f"(?:{s})"
     return s
 
-def _separators_to_re(separators, as_bytes, separate=False, keep=False):
+
+@functools.lru_cache(re._MAXCACHE)
+def __separators_to_re(separators, as_bytes, separate=False, keep=False):
     if as_bytes:
         pipe = b'|'
         separate_start = b'(?:'
@@ -325,6 +333,14 @@ def _separators_to_re(separators, as_bytes, separate=False, keep=False):
     # print(f"  P3 {pattern!r}")
     return pattern
 
+def _separators_to_re(separators, as_bytes, separate=False, keep=False):
+    try:
+        hash(separators)
+    except TypeError:
+        separators = tuple(separators)
+    return __separators_to_re(separators, as_bytes, separate=separate, keep=keep)
+
+
 @_export
 def multistrip(s, separators, left=True, right=True):
     """
@@ -358,8 +374,9 @@ def multistrip(s, separators, left=True, right=True):
         tail = b'$'
         if isinstance(separators, bytes):
             # not iterable of bytes, literally a bytes string.
-            # split it ourselves.
-            separators = [separators[i:i+1] for i in range(len(separators))]
+            # split it ourselves.  otherwise, _separators_to_re will
+            # iterate over it, and... oops! it gets bytes!
+            separators = tuple(separators[i:i+1] for i in range(len(separators)))
     else:
         head = '^'
         tail = '$'
@@ -422,6 +439,8 @@ def multisplit(s, separators=None, *,
             Discard the separators.
         true (apart from ALTERNATING or AS_PAIRS)
             Append the separators to the end of the split strings.
+            You can recreate the original string by passing the
+            list returned in to ''.join.
         ALTERNATING
             Yield alternating strings in the output: strings consisting
             of separators, alternating with strings consisting of
@@ -430,11 +449,14 @@ def multisplit(s, separators=None, *,
             may be empty; if "separate" is false, separator strings will
             contain one or more separators, and non-separator strings
             will never be empty.
+            You can recreate the original string by passing the
+            list returned in to ''.join.
         AS_PAIRS
             Yield 2-tuples containing a non-separator string and its
-            subsequent separator string.  Either string may be empty,
-            but both strings will never be empty at the same time,
-            unless s was empty.
+            subsequent separator string.  Either string may be empty;
+            the separator string in the last 2-tuple will always be
+            empty, and if `s` ends with a separator string, *both*
+            strings in the final 2-tuple will be empty.
 
     "separate" indicates whether multisplit should consider adjacent
     separator strings in s as one separator or as multiple separators
@@ -465,15 +487,15 @@ def multisplit(s, separators=None, *,
         true
             Strip separators from the beginning and end of s
             (a la str.strip).
-        STR_STRIP (currently unimplemented, equivalent to true)
-            Behave like str.strip.  Strip from the beginning
-            and end of s, unless maxsplit is nonzero and the
-            entire string is not split.  If splitting stops due
-            to maxsplit before the entire string is split, and
-            reverse is false, don't strip the end of the string.
-            If splitting stops due to maxsplit before the entire
-            string is split, and reverse is true, don't strip
-            the beginning of the string.
+        STR_SPLIT (currently unimplemented, equivalent to true)
+            Strip the way str.split does when splitting on whitespace.
+            Strip from the beginning and end of s, unless maxsplit
+            is nonzero and the entire string is not split.  If
+            splitting stops due to maxsplit before the entire string
+            is split, and reverse is false, don't strip the end of
+            the string. If splitting stops due to maxsplit before
+            the entire string is split, and reverse is true, don't
+            strip the beginning of the string.
 
     "maxsplit" should be either an integer or None.  If maxsplit is an
     integer greater than -1, multisplit will split s no more than
@@ -491,11 +513,12 @@ def multisplit(s, separators=None, *,
     as_bytes = isinstance(s, bytes)
     if separators is None:
         separators = ascii_whitespace if as_bytes else whitespace
+
     if as_bytes:
         if isinstance(separators, bytes):
             # not iterable of bytes, literally a bytes string.
             # split it ourselves.
-            separators = [separators[i:i+1] for i in range(len(separators))]
+            separators = tuple(separators[i:i+1] for i in range(len(separators)))
         separators_type = bytes
         empty = b''
     else:
@@ -578,7 +601,8 @@ def multisplit(s, separators=None, *,
             yield o
         return
 
-    # from here on out, we only deal with keep=True
+    # from here on out, we're 'keep'-ing the separator strings.
+    # (we're returning the separator strings in one form or another.)
 
     if keep == ALTERNATING:
         for o in l:
@@ -594,8 +618,7 @@ def multisplit(s, separators=None, *,
             previous = o
             continue
         if keep == AS_PAIRS:
-            if previous or o:
-                yield (previous, o)
+            yield (previous, o)
         else:
             yield previous + o
         previous = None
@@ -659,18 +682,18 @@ def multipartition(s, separators, count=1, *, reverse=False, separate=True):
     return tuple(result)
 
 
-_invalid_state = 0
-_in_word = 1
-_after_whitespace = 2
-_after_whitespace_then_apostrophe_or_quote_mark = 3
-_after_whitespace_then_D_or_O = 4
-_after_whitespace_then_D_or_O_then_apostrophe = 5
+_invalid_state = "_invalid_state"
+_in_word = "_in_word"
+_after_whitespace = "_after_whitespace"
+_after_whitespace_then_apostrophe_or_double_quote = "_after_whitespace_then_apostrophe_or_double_quote"
+_after_whitespace_then_D_or_O = "_after_whitespace_then_D_or_O"
+_after_whitespace_then_D_or_O_then_apostrophe = "_after_whitespace_then_D_or_O_then_apostrophe"
 
 @_export
-def gently_title(s):
+def gently_title(s, *, apostrophes=None, double_quotes=None):
     """
     Uppercase the first character of every word in s.
-    Leave the other letters alone.
+    Leave the other letters alone.  s should be str or bytes.
 
     (For the purposes of this algorithm, words are
     any blob of non-whitespace characters.)
@@ -699,34 +722,68 @@ def gently_title(s):
     quote marks is considered one quote mark for
     the purposes of capitalization.
 
-    Each of these Unicode code points is considered
-    an apostrophe:
+    If specified, apostrophes should be a str
+    or bytes object containing characters that
+    should be considered apostrophes.  If apostrophes
+    is false, and s is bytes, apostrophes is set to "'".
+    If apostrophes is false and s is str, apostrophes
+    is set to a string containing these Unicode
+    apostrophe code points:
         '‘’‚‛
 
-    Each of these Unicode code points is considered
-    a quote mark:
+    If specified, double_quotes should be a str
+    or bytes object containing characters that
+    should be considered double-quote characters.
+    If double_quotes is false, and s is bytes,
+    double_quotes is set to "'".
+    If double_quotes is false, and s is str, double_quotes
+    is set to a string containing these Unicode double quote
+    code points:
         "“”„‟«»‹›
     """
+    as_bytes = isinstance(s, bytes)
+    if as_bytes:
+        empty = b""
+        if not apostrophes:
+            apostrophes = b"'"
+        if not double_quotes:
+            double_quotes = b'"'
+        d_and_o = b'DO'
+        characters_in_s = (s[i:i+1] for i in range(len(s)))
+    else:
+        empty = ""
+        if not apostrophes:
+            apostrophes = "'‘’‚‛"
+        if not double_quotes:
+            double_quotes = '"“”„‟«»‹›'
+        d_and_o = 'DO'
+        characters_in_s = iter(s)
+
+    # ooh, fancy!
+    _is_apostrophe = apostrophes.__contains__
+    _is_double_quote = double_quotes.__contains__
+    _is_d_or_o = d_and_o.__contains__
+
     result = []
     state = _after_whitespace
-    for c in s:
+    for c in characters_in_s:
         is_space = c.isspace()
-        is_apostrophe = c in "'‘’‚‛"
-        is_quote_mark = c in '"“”„‟«»‹›'
+        is_apostrophe = _is_apostrophe(c)
+        is_double_quote = _is_double_quote(c)
         if state == _in_word:
             if is_space:
                 state = _after_whitespace
         elif state == _after_whitespace:
             if not is_space:
                 c = c.upper()
-                if (is_apostrophe or is_quote_mark):
-                    state = _after_whitespace_then_apostrophe_or_quote_mark
-                elif c in "DO":
+                if (is_apostrophe or is_double_quote):
+                    state = _after_whitespace_then_apostrophe_or_double_quote
+                elif _is_d_or_o(c):
                     state = _after_whitespace_then_D_or_O
                 else:
                     state = _in_word
-        elif state == _after_whitespace_then_apostrophe_or_quote_mark:
-            if not (is_apostrophe or is_quote_mark):
+        elif state == _after_whitespace_then_apostrophe_or_double_quote:
+            if not (is_apostrophe or is_double_quote):
                 c = c.upper()
                 state = _in_word
         elif state == _after_whitespace_then_D_or_O:
@@ -738,7 +795,7 @@ def gently_title(s):
             c = c.upper()
             state = _in_word
         result.append(c)
-    return "".join(result)
+    return empty.join(result)
 
 
 @_export
@@ -749,10 +806,14 @@ def normalize_whitespace(s, separators=None, replacement=None):
     By default turns all runs of consecutive whitespace
     characters into a single space character.
 
-    "s" may be str or bytes.
-    "separators" should be an iterable of either str or bytes,
-    matching "s".
-    "replacement" should be str or bytes, also matching "s".
+    s may be str or bytes.
+    separators should be an iterable of either str or bytes objects,
+    matching s.
+    replacement should be either a str or bytes object,
+    also matching s, or None (the default).
+    If replacement is None, normalize_whitespace will use
+    a replacement string consisting of a single space character,
+    either str or bytes as appropriate.
 
     Leading or trailing runs of separator characters will
     be replaced with the replacement string, e.g.:
@@ -802,7 +863,7 @@ def normalize_whitespace(s, separators=None, replacement=None):
 
 
 @_export
-def split_quoted_strings(s, quotes=('"', "'"), *, triple_quotes=True, backslash='\\'):
+def split_quoted_strings(s, quotes=None, *, triple_quotes=True, backslash=None):
     """
     Splits s into quoted and unquoted segments.  Returns an iterator yielding
     2-tuples:
@@ -819,7 +880,24 @@ def split_quoted_strings(s, quotes=('"', "'"), *, triple_quotes=True, backslash=
     If backslash is a character, this character will quoting characters inside
     a quoted string, like the backslash character inside strings in Python.
     """
-    i = PushbackIterator(s)
+    as_bytes = isinstance(s, bytes)
+    if as_bytes:
+        empty = b''
+        if quotes is None:
+            quotes=(b'"', b"'")
+        if backslash is None:
+            backslash = b'\\'
+        i = (s[i:i+1] for i in range(len(s)))
+    else:
+        empty = ''
+        if quotes is None:
+            quotes=('"', "'")
+        if backslash is None:
+            backslash = '\\'
+        i = s
+    empty_join = empty.join
+
+    i = PushbackIterator(i)
     in_quote = None
     in_triple_quote = False
     in_backslash = False
@@ -835,7 +913,7 @@ def split_quoted_strings(s, quotes=('"', "'"), *, triple_quotes=True, backslash=
                 # triple quotes are off, encountered quote.
                 # flush unquoted string, start quoted string.
                 if text:
-                    yield False, "".join(text)
+                    yield False, empty_join(text)
                     text.clear()
                 text.append(c)
                 in_quote = c
@@ -846,7 +924,7 @@ def split_quoted_strings(s, quotes=('"', "'"), *, triple_quotes=True, backslash=
                 # only a single quote mark. flush unquoted string, start quoted string.
                 i.push(c2)
                 if text:
-                    yield False, "".join(text)
+                    yield False, empty_join(text)
                     text.clear()
                 text.append(c)
                 in_quote = c
@@ -856,7 +934,7 @@ def split_quoted_strings(s, quotes=('"', "'"), *, triple_quotes=True, backslash=
                 # two quotes in a row, but not three.
                 # flush unquoted string, emit empty quoted string.
                 if text:
-                    yield False, "".join(text)
+                    yield False, empty_join(text)
                     text.clear()
                 yield True, c*2
                 if c3 is not None:
@@ -865,7 +943,7 @@ def split_quoted_strings(s, quotes=('"', "'"), *, triple_quotes=True, backslash=
             # triple quoted string.
             # flush unquoted string, start triple quoted string.
             if text:
-                yield False, "".join(text)
+                yield False, empty_join(text)
                 text.clear()
             text.append(c*3)
             in_quote = c
@@ -897,7 +975,7 @@ def split_quoted_strings(s, quotes=('"', "'"), *, triple_quotes=True, backslash=
             # finish and emit the quoted string,
             # and return to unquoted mode.
             text.append(c)
-            yield True, "".join(text)
+            yield True, empty_join(text)
             text.clear()
             in_quote = False
             continue
@@ -911,7 +989,7 @@ def split_quoted_strings(s, quotes=('"', "'"), *, triple_quotes=True, backslash=
             # finish and emit the triple-quoted string,
             # and return to unquoted mode.
             text.append(c*3)
-            yield True, "".join(text)
+            yield True, empty_join(text)
             text.clear()
             in_triple_quote = in_quote = False
             continue
@@ -927,7 +1005,7 @@ def split_quoted_strings(s, quotes=('"', "'"), *, triple_quotes=True, backslash=
     # flush the remainder of the string we were building,
     # in whatever condition it's in.
     if text:
-        yield in_quote, "".join(text)
+        yield in_quote, empty_join(text)
 
 
 @_export
@@ -1105,7 +1183,13 @@ def lines_filter_comment_lines(li, comment_separators):
     """
     if not comment_separators:
         raise ValueError("illegal comment_separators")
-    as_bytes = isinstance(comment_separators[0], bytes)
+
+    if isinstance(comment_separators, bytes):
+        comment_separators = tuple(comment_separators[i:i+1] for i in range(len(comment_separators)))
+        as_bytes = True
+    else:
+        as_bytes = isinstance(comment_separators[0], bytes)
+
     comment_pattern = _separators_to_re(tuple(comment_separators), as_bytes, separate=False, keep=False)
     comment_re = re.compile(comment_pattern)
     for info, line in li:
@@ -1114,6 +1198,73 @@ def lines_filter_comment_lines(li, comment_separators):
             continue
         yield (info, line)
 
+
+@_export
+def lines_filter_contains(li, s, *, invert=False):
+    """
+    A lines modifier function.  Only yields lines
+    that contain s.  (Filters out lines that
+    don't contain s.)
+
+    If invert is true, returns the opposite--
+    filters out lines that contain s.
+
+    Composable with all the lines_ functions from the big.text module.
+    """
+    if invert:
+        for t in li:
+            if not s in t[1]:
+                yield t
+        return
+    for t in li:
+        if s in t[1]:
+            yield t
+
+@_export
+def lines_filter_grep(li, pattern, *, invert=False, flags=0):
+    """
+    A lines modifier function.  Only yields lines
+    that match the regular expression pattern.
+    (Filters out lines that don't match pattern.)
+
+    pattern can be str, bytes, or an re.Pattern object.
+    If pattern is not an re.Pattern object, it's compiled
+    with re.compile(pattern, flags=flags).
+
+    If invert is true, returns the opposite--
+    filters out lines that match pattern.
+
+    Composable with all the lines_ functions from the big.text module.
+    """
+    if not isinstance(pattern, re.Pattern):
+        pattern = re.compile(pattern, flags=flags)
+    search = pattern.search
+    if invert:
+        for t in li:
+            if not search(t[1]):
+                yield t
+        return
+    for t in li:
+        if search(t[1]):
+            yield t
+
+@_export
+def lines_sort(li, *, reverse=False):
+    """
+    A lines modifier function.  Sorts all
+    input lines before yielding them.
+
+    Lines are sorted lexicographically,
+    from lowest to highest.
+    If reverse is true, lines are sorted
+    from highest to lowest.
+
+    Composable with all the lines_ functions from the big.text module.
+    """
+    lines = list(li)
+    lines.sort(key=lambda t:t[1], reverse=reverse)
+    yield from iter(lines)
+
 @_export
 def lines_strip_comments(li, comment_separators, *, quotes=('"', "'"), backslash='\\', rstrip=True, triple_quotes=True):
     """
@@ -1156,7 +1307,18 @@ def lines_strip_comments(li, comment_separators, *, quotes=('"', "'"), backslash
     """
     if not comment_separators:
         raise ValueError("illegal comment_separators")
-    as_bytes = isinstance(comment_separators[0], bytes)
+
+    if isinstance(comment_separators, bytes):
+        comment_separators = tuple(comment_separators[i:i+1] for i in range(len(comment_separators)))
+        as_bytes = True
+    else:
+        as_bytes = isinstance(comment_separators[0], bytes)
+
+    if as_bytes:
+        empty = b''
+    else:
+        empty = ''
+    empty_join = empty.join
 
     comment_pattern = _separators_to_re(comment_separators, as_bytes=as_bytes, separate=True, keep=True)
     re_comment = re.compile(comment_pattern)
@@ -1187,8 +1349,8 @@ def lines_strip_comments(li, comment_separators, *, quotes=('"', "'"), backslash
             append(leading)
             comment = fields[1:]
 
-        info.comment = "".join(comment)
-        line = "".join(segments)
+        info.comment = empty_join(comment)
+        line = empty_join(segments)
         yield (info, line)
 
 @_export
