@@ -70,6 +70,62 @@ def printable_separators(separators):
 
 class BigTextTests(unittest.TestCase):
 
+    def test_whitespace_and_newlines(self):
+        # ensure that big.whitespace and big.newlines
+        # correctly matches the list of characters that
+        # Python considers whitespace / newlines.
+
+        # Python whitespace only considers individual
+        # whitespace charcters, and doesn't include the
+        # DOS end-of-line sequence '\r\n'.  so what we're
+        # going to produce in the below list is technically
+        # the "without DOS" versions.
+        python_whitespace_without_dos = []
+        python_newlines_without_dos = []
+
+        # technically we don't need to check the surrogate pair characters.
+        # but it's faster to leave them in.
+        unicode_code_points = 2**16 + 2**20
+
+        for i in range(unicode_code_points):
+            c = chr(i)
+            s = f"a{c}b"
+            if len(s.split()) == 2:
+                python_whitespace_without_dos.append(c)
+                if len(s.splitlines()) == 2:
+                    python_newlines_without_dos.append(c)
+
+        self.assertEqual(set(big.whitespace_without_dos), set(python_whitespace_without_dos))
+        self.assertEqual(set(big.newlines_without_dos), set(python_newlines_without_dos))
+
+        python_whitespace = list(python_whitespace_without_dos)
+        python_whitespace.append("\r\n")
+        python_newlines = list(python_newlines_without_dos)
+        python_newlines.append("\r\n")
+        self.assertEqual(set(big.whitespace), set(python_whitespace))
+        self.assertEqual(set(big.newlines), set(python_newlines))
+
+        # now test the utf-8 and ascii variants!
+
+        python_ascii_whitespace = big.text._cheap_encode_iterable_of_strings(python_whitespace, 'ascii')
+        self.assertEqual(set(big.ascii_whitespace), set(python_ascii_whitespace))
+        python_ascii_whitespace_without_dos = big.text._cheap_encode_iterable_of_strings(python_whitespace_without_dos, 'ascii')
+        self.assertEqual(set(big.ascii_whitespace_without_dos), set(python_ascii_whitespace_without_dos))
+        python_ascii_newlines = big.text._cheap_encode_iterable_of_strings(python_newlines, 'ascii')
+        self.assertEqual(set(big.ascii_newlines), set(python_ascii_newlines))
+        python_ascii_newlines_without_dos = big.text._cheap_encode_iterable_of_strings(python_newlines_without_dos, 'ascii')
+        self.assertEqual(set(big.ascii_newlines_without_dos), set(python_ascii_newlines_without_dos))
+
+        python_utf8_whitespace = big.text._cheap_encode_iterable_of_strings(python_whitespace, 'utf8')
+        self.assertEqual(set(big.utf8_whitespace), set(python_utf8_whitespace))
+        python_utf8_whitespace_without_dos = big.text._cheap_encode_iterable_of_strings(python_whitespace_without_dos, 'utf8')
+        self.assertEqual(set(big.utf8_whitespace_without_dos), set(python_utf8_whitespace_without_dos))
+        python_utf8_newlines = big.text._cheap_encode_iterable_of_strings(python_newlines, 'utf8')
+        self.assertEqual(set(big.utf8_newlines), set(python_utf8_newlines))
+        python_utf8_newlines_without_dos = big.text._cheap_encode_iterable_of_strings(python_newlines_without_dos, 'utf8')
+        self.assertEqual(set(big.utf8_newlines_without_dos), set(python_utf8_newlines_without_dos))
+
+
     def test_re_partition(self):
         def group0(re_partition_result):
             result = []
@@ -87,6 +143,8 @@ class BigTextTests(unittest.TestCase):
                 c(("abc", "123", "def456ghi")) )
             self.assertEqual(group0(big.re_rpartition(s, pattern)),
                 c(("abc123def", "456", "ghi")) )
+            self.assertEqual(group0(big.re_partition(s, pattern, reverse=True)),
+                c(("abc123def", "456", "ghi")) )
 
             s = c("abc12345def67890ghi")
             self.assertEqual(group0(big.re_partition(s, pattern)),
@@ -139,6 +197,8 @@ class BigTextTests(unittest.TestCase):
             def test_re_rpartition(s, pattern, count, expected):
                 self.assertEqual(group0(big.re_rpartition(s, pattern, count)),
                     expected )
+                self.assertEqual(group0(big.re_partition(s, pattern, count, reverse=True)),
+                    expected )
 
             test_re_rpartition("a:b:c:d", ":", 0, ("a:b:c:d",))
             test_re_rpartition("a:b:c:d", ":", 1, ("a:b:c", ':' ,"d"))
@@ -248,8 +308,9 @@ class BigTextTests(unittest.TestCase):
             # test greedy separators
             self.assertEqual(multisplit(c('-abcde-abc-a-abc-abcde-'), c(['a', 'abc', 'abcde'])), c(['-', '-', '-', '-', '-', '-']))
 
-            # regression test: don't ever end with a tuple containing two empty strings
-            self.assertEqual(multisplit(c('\na\nb\nc\n'), c(('\n',)), keep=big.AS_PAIRS, strip=False), c([ ('', '\n'), ('a', '\n'), ('b', '\n'), ('c', '\n') ]))
+            # regression test: *YES*, if the string you're splitting ends with a separator,
+            # and keep=big.AS_PAIRS, the result ends with a tuple containing two empty strings.
+            self.assertEqual(multisplit(c('\na\nb\nc\n'), c(('\n',)), keep=big.AS_PAIRS, strip=False), c([ ('', '\n'), ('a', '\n'), ('b', '\n'), ('c', '\n'), ('', '') ]))
 
         with self.assertRaises(TypeError):
             multisplit('s', 3.1415)
@@ -657,8 +718,7 @@ class BigTextTests(unittest.TestCase):
                                                     if waiting is None:
                                                         waiting = s
                                                         return
-                                                    if waiting or s:
-                                                        new_expected.append((waiting, s))
+                                                    new_expected.append((waiting, s))
                                                     waiting = None
                                                 for s in expected:
                                                     append(s)
@@ -1096,9 +1156,16 @@ class BigTextTests(unittest.TestCase):
         )
 
     def test_gently_title(self):
-        def test(s, expected):
-            result = big.gently_title(s)
+        def test(s, expected, test_ascii=True, apostrophes=None, double_quotes=None):
+            result = big.gently_title(s, apostrophes=apostrophes, double_quotes=double_quotes)
             self.assertEqual(result, expected)
+            if test_ascii:
+                if apostrophes:
+                    apostrophes = apostrophes.encode('ascii')
+                if double_quotes:
+                    double_quotes = double_quotes.encode('ascii')
+                result = big.gently_title(s.encode('ascii'), apostrophes=apostrophes, double_quotes=double_quotes)
+                self.assertEqual(result, expected.encode('ascii'))
 
         test("", "")
         test("abcde fgh", "Abcde Fgh")
@@ -1110,7 +1177,9 @@ class BigTextTests(unittest.TestCase):
         test("everybody's thinking they couldn't've had a v-8", "Everybody's Thinking They Couldn't've Had A V-8")
         test("don't come home if you don't get 1st", "Don't Come Home If You Don't Get 1st")
         test("""i said "no, i didn't", you idiot""", """I Said "No, I Didn't", You Idiot""")
-        test('multiple «"“quote marks”"»', 'Multiple «"“Quote Marks”"»')
+        test('multiple «"“quote marks”"»', 'Multiple «"“Quote Marks”"»', test_ascii=False)
+
+        test("""i said ZdonXt touch that, oXconnell!Z, you 2nd rate idiot!""", """I Said ZDonXt Touch That, OXConnell!Z, You 2nd Rate Idiot!""", apostrophes='X', double_quotes='Z')
 
     def test_normalize_whitespace(self):
         def test(s, expected, *, separators=None, replacement=" "):
@@ -1148,6 +1217,9 @@ class BigTextTests(unittest.TestCase):
             got = list(big.split_quoted_strings(s, **kwargs))
             self.assertEqual(got, expected)
 
+            got = list(big.split_quoted_strings(s.encode('ascii'), **kwargs))
+            self.assertEqual(got, [(b, s.encode('ascii')) for b, s in expected])
+
         test("""hey there "this is quoted" an empty quote: '' this is not quoted 'this is more quoted' "here's quoting a quote mark: \\" wow!" this is working!""",
             [
                 (False, 'hey there '),
@@ -1188,6 +1260,10 @@ class BigTextTests(unittest.TestCase):
 
         LI = big.LineInfo
 
+        def L(line, line_number, column_number=1, **kwargs):
+            info = big.LineInfo(line, line_number, column_number, **kwargs)
+            return (info, line)
+
         test(big.lines("a\nb\nc\nd\ne\n"),
             [
             (LI('a', 1, 1), 'a'),
@@ -1223,6 +1299,125 @@ class BigTextTests(unittest.TestCase):
             (LI('',                    9, 1), ''),
             ])
 
+        test(big.lines_filter_comment_lines(big.lines(b"a\n# ignored\n c"), b'#'),
+            [
+            L(b'a', 1),
+            L(b' c', 3),
+            ]
+            )
+
+        test(big.lines_filter_contains(big.lines("""
+hello yolks
+what do you have to say, champ?
+i like eggs.
+they don't have to be fancy.
+simple scrambled eggs are just fine.
+neggatory!
+whoops, I meant, negatory.
+"""[1:]), "egg"),
+            [
+                (LI('i like eggs.', 3, 1), 'i like eggs.'),
+                (LI('simple scrambled eggs are just fine.', 5, 1), 'simple scrambled eggs are just fine.'),
+                (LI('neggatory!', 6, 1), 'neggatory!'),
+            ]
+            )
+
+        test(big.lines_filter_contains(big.lines("""
+hello yolks
+what do you have to say, champ?
+i like eggs.
+they don't have to be fancy.
+simple scrambled eggs are just fine.
+neggatory!
+whoops, I meant, negatory.
+"""[1:]), "egg", invert=True),
+            [
+                (LI('hello yolks', 1, 1), 'hello yolks'),
+                (LI('what do you have to say, champ?', 2, 1), 'what do you have to say, champ?'),
+                (LI("they don't have to be fancy.", 4, 1), "they don't have to be fancy."),
+                (LI('whoops, I meant, negatory.', 7, 1), 'whoops, I meant, negatory.'),
+                (LI('', 8, 1), ''),
+            ]
+            )
+
+        test(big.lines_filter_grep(big.lines("""
+hello yolks
+what do you have to say, champ?
+i like eggs.
+they don't have to be fancy.
+simple scrambled eggs are just fine.
+neggatory!
+whoops, I meant, negatory.
+"""[1:]), "eg+"),
+            [
+                (LI('i like eggs.', 3, 1), 'i like eggs.'),
+                (LI('simple scrambled eggs are just fine.', 5, 1), 'simple scrambled eggs are just fine.'),
+                (LI('neggatory!', 6, 1), 'neggatory!'),
+                (LI('whoops, I meant, negatory.', 7, 1), 'whoops, I meant, negatory.'),
+            ]
+            )
+
+        test(big.lines_filter_grep(big.lines("""
+hello yolks
+what do you have to say, champ?
+i like eggs.
+they don't have to be fancy.
+simple scrambled eggs are just fine.
+neggatory!
+whoops, I meant, negatory.
+"""[1:]), "eg+", invert=True),
+            [
+                (LI('hello yolks', 1, 1), 'hello yolks'),
+                (LI('what do you have to say, champ?', 2, 1), 'what do you have to say, champ?'),
+                (LI("they don't have to be fancy.", 4, 1), "they don't have to be fancy."),
+                (LI('', 8, 1), ''),
+            ]
+            )
+
+        test(big.lines_sort(big.lines("""
+cormorant
+firefox
+alligator
+diplodocus
+elephant
+giraffe
+barracuda
+hummingbird
+"""[1:-1])),
+            [
+                L('alligator', 3),
+                L('barracuda', 7),
+                L('cormorant', 1),
+                L('diplodocus', 4),
+                L('elephant', 5),
+                L('firefox', 2),
+                L('giraffe', 6),
+                L('hummingbird', 8),
+            ]
+            )
+
+        test(big.lines_sort(big.lines("""
+cormorant
+firefox
+alligator
+diplodocus
+elephant
+giraffe
+barracuda
+hummingbird
+"""[1:-1]), reverse=True),
+            [
+                L('hummingbird', 8),
+                L('giraffe', 6),
+                L('firefox', 2),
+                L('elephant', 5),
+                L('diplodocus', 4),
+                L('cormorant', 1),
+                L('barracuda', 7),
+                L('alligator', 3),
+            ]
+            )
+
         test(big.lines_rstrip(big.lines(
 "    a = b  \n"
 "    c = d     \n"
@@ -1309,6 +1504,13 @@ for x in range(5): # this is a comment
         with self.assertRaises(ValueError):
             test(big.lines_strip_comments(big.lines("a\nb\n"), None), [])
 
+        test(big.lines_strip_comments(big.lines(b"a\nb# ignored\n c"), b'#'),
+            [
+            L(b'a', 1, comment=b''),
+            (LI(b'b# ignored', 2, 1, comment=b'# ignored'), b'b'),
+            L(b' c', 3, comment=b''),
+            ]
+            )
 
 
         test(big.lines_filter_empty_lines(big.lines_filter_comment_lines(big.lines_strip(big.lines(
